// MachPCBDrill, NC drill/routing program for Mach3
// Copyright (C) 201, Frank Tkalcevic, www.franksworkshop.com

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "StdAfx.h"
#include "NCDrillFile.h"



CNCDrillFile::CNCDrillFile(void)
    : m_bMetric( true ),
    m_bLeadingZeroes( false ),
    m_nFormat( 1 ),
    m_bDrilling( true ),	// default mode drill
    m_bPlunged( false )
{
}

CNCDrillFile::~CNCDrillFile(void)
{
    std::list<NCCoord*>::iterator itCoord;
    for ( itCoord = m_Coords.begin(); itCoord != m_Coords.end(); itCoord++ )
	delete *itCoord;
    m_Coords.clear();

    std::map<int,NCTool*>::iterator itTool;
    for ( itTool = m_Tools.begin(); itTool != m_Tools.end(); itTool++ )
	delete itTool->second;
    m_Tools.clear();
}

void CNCDrillFile::ProcessZeroes( CString &sLine )
{
    if (sLine.GetLength() == 0)
    {
	m_bLeadingZeroes = true;
    }
    else if (sLine.GetLength() >= 3 && sLine.Left(3) == ",LZ")
    {
	m_bLeadingZeroes = true;
    }
    else if (sLine.GetLength() >= 3 && sLine.Left(3) == ",TZ")
    {
	m_bLeadingZeroes = false;
    }

}

void CNCDrillFile::ProcessToolDefintion( CString &sLine )
{
    CAtlRegExp<> r;
    REParseError status = r.Parse( "^{\\d+}[^C]*C{[.0-9]+}");

    CAtlREMatchContext<> mc;

    if ( r.Match(sLine,&mc) )
    {
	int nGroups = mc.m_uNumGroups;

	const CAtlREMatchContext<>::RECHAR* szStart = 0;
	const CAtlREMatchContext<>::RECHAR* szEnd = 0;

	mc.GetMatch(0, &szStart, &szEnd);
	ptrdiff_t nLength = szEnd - szStart;
	CString sId( szStart, (int)nLength );

	mc.GetMatch(1, &szStart, &szEnd);
	nLength = szEnd - szStart;
	CString sSize( szStart, (int)nLength );

	int nId = atoi( sId );
	double nSize = atof( sSize );

	NCTool *t = new NCTool(nId, nSize);
	m_Tools[nId] = t;
    }
    else
    {
	theApp.LogMessage( "Line %d: Failed to parse line '%s'", m_nLineCount, (LPCSTR)sLine);
    }
}


double CNCDrillFile::MakeDouble( CString &sLine )
{
    CString sCoord;
    if (m_bLeadingZeroes)
    {
	if (m_bMetric)
	{
	    while ( sLine.GetLength() < 4 )
		sLine += "0";
	    sCoord = sLine.Mid(0, 4) + "." + sLine.Mid(4);
	}
	else
	{
	    while ( sLine.GetLength() < 2 )
		sLine += "0";
	    sCoord = sLine.Mid(0, 2) + "." + sLine.Mid(2);
	}
    }
    return atof(sCoord);
}

void CNCDrillFile::GetXY( CString sLine, double &x, double &y )
{
    x = m_dblLastX;
    y = m_dblLastY;

    while ( !sLine.IsEmpty() )
    {
	char cAxis = sLine[0];
	sLine = sLine.Mid(1);
	int nEndIndex = sLine.FindOneOf("XY");
	CString sCoord;
	if ( nEndIndex >= 0 )
	{
	    sCoord = sLine.Mid(0,nEndIndex);
	    sLine = sLine.Mid( nEndIndex );
	}
	else
	{
	    sCoord = sLine;
	    sLine.Empty();
	}

	double dblCoord = MakeDouble(sCoord);

	if (cAxis == 'X')
	    x = m_dblLastX = dblCoord;
	else if (cAxis == 'Y')
	    y = m_dblLastY = dblCoord;
	else
	{
	    theApp.LogMessage( "Line %d: Unknown coordinate axis '%c'", m_nLineCount, cAxis);
	}
    }
}

void CNCDrillFile::ProcessHole( CString &sLine )
{
    double x, y;
    GetXY( sLine, x, y );
    m_Coords.push_back( new NCCoord(m_ActiveTool, x, y, m_nLineCount));
}
void CNCDrillFile::ProcessToolChange( CString &sLine )
{
    int nId = atoi(sLine);
    std::map<int,NCTool*>::iterator itFind = m_Tools.find(nId);
    if ( itFind == m_Tools.end() )
    {
	theApp.LogMessage( "Line %d: Unknown tool '%d'", m_nLineCount, nId);
    }
    else
	m_ActiveTool = itFind->second;
}

void CNCDrillFile::SetFormat( CString &s )
{
    m_nFormat = atoi( s );
}


void CNCDrillFile::ProcessCutterCompensation(CString &sLine )
{
}

void CNCDrillFile::ProcessRoute(CString &sLine )
{
    double x, y;
    GetXY( sLine, x, y );
    m_Coords.push_back( new NCRoute(m_ActiveTool, x, y, m_nLineCount, m_bPlunged ) );
}

void CNCDrillFile::ProcessGCode(CString sLine )
{
    CString sCode = sLine.Mid(1,2);
    sLine = sLine.Mid(3);

    if ( sCode == "00" )
    {
	//G00[XY] - Route mode
	m_bDrilling = false;
	if ( !sLine.IsEmpty() )
	    ProcessRoute( sLine );
    }
    else if ( sCode == "05" )
    {
	//G05[XY] - drill mode
	m_bDrilling = true;
	if ( !sLine.IsEmpty() )
	    ProcessHole( sLine );
    }
    else if ( sCode == "01" )
    {
	//G01[XY] -linear move
	if ( !sLine.IsEmpty() )
	    ProcessRoute( sLine );
    }
    else if ( sCode == "40" )
    {
	//G40 - cutter compensation, off
    }
    else if ( sCode == "41" )
    {
	//G41 - cutter compensation, left
    }
    else if ( sCode == "42" )
    {
	//G42 - cutter compensation, right
    }

}

void CNCDrillFile::ProcessMCode(CString &sLine )
{
    if ( sLine.Mid(0,3) == "M15" )
	//M15 - plunge
	m_bPlunged = true;
    else if ( sLine.Mid(0,3) == "M16" )
	//M16 - retract
	m_bPlunged = false;

}

void CNCDrillFile::Load(CString sFilename)
{
    CStdioFile file;
    if ( !file.Open( sFilename, CFile::modeRead | CFile::shareDenyWrite | CFile::typeText ) )
    {
	ASSERT( FALSE );
	return;
    }

    enum EReadState
    {
	ReadingHeader,
	ReadingBody,
	Done
    } eState = ReadingBody;

    m_ActiveTool = 0;
    m_nLineCount = 0;
    CString sLine;
    while (file.ReadString(sLine))
    {
	sLine = sLine.Trim();

	switch (eState)
	{
	case ReadingHeader:
	    if (sLine.GetLength() >= 6 && sLine.Left(6) == "METRIC")
	    {
		m_bMetric = true;
		ProcessZeroes( sLine.Mid(6) );
	    }
	    else if (sLine.GetLength() >= 4 && sLine.Left(4) == "INCH")
	    {
		m_bMetric = false;
		ProcessZeroes( sLine.Mid(4) );
	    }
	    else if (sLine.GetLength() >= 1 && sLine.Left(1) == ";")
	    {
		// comment
	    }
	    else if (sLine.GetLength() >= 2 && sLine.Left(1) == "T")
	    {
		// Tool definition
		ProcessToolDefintion(sLine.Mid(1));
	    }
	    else if (sLine.GetLength() >= 1 && sLine.Left(1) == "%")
	    {
		// End of header
		eState = ReadingBody;
	    }
	    else if (sLine.GetLength() >= 3 && sLine.Left(3) == "M95")
	    {
		// End of header
		eState = ReadingBody;
	    }
	    else if (sLine.GetLength() >= 5 && sLine.Left(5) == "FMAT,")
	    {
		// End of header
		SetFormat( sLine.Mid(5) );
	    }
	    else if (sLine.GetLength() >= 7 && sLine.Left(7) == "DETECT,")
	    {
		// Broken tool detection
	    }
	    else if (sLine.GetLength() >= 2 && sLine.Left(1) == "CP")
	    {
		// Cutter Compensation
		ProcessCutterCompensation(sLine.Mid(2));
	    }
	    break;

	case ReadingBody:
	    if (sLine.GetLength() >= 3 && sLine.Left(3) == "M48")
		eState = ReadingHeader;
	    else if (sLine.GetLength() >= 3 && sLine.Left(3) == "M30")
	    {
		eState = Done;
	    }
	    else if (sLine.GetLength() >= 1 && (sLine.Left(1) == "X" || sLine.Left(1) == "Y") )
	    {
		if ( m_bDrilling )
		    ProcessHole(sLine);
		else
		    ProcessRoute(sLine);
	    }
	    else if (sLine.GetLength() >= 2 && sLine.Left(1) == "T")
	    {
		// Tool definition
		ProcessToolChange(sLine.Mid(1));
	    }
	    else if (sLine.GetLength() >= 1 && sLine.Left(1) == "G")
	    {
		ProcessGCode( sLine );
	    }
	    else if (sLine.GetLength() >= 1 && sLine.Left(1) == "M")
	    {
		ProcessMCode( sLine );
	    }
	    break;

	case Done:
	    break;
	}

	m_nLineCount++;
    }

    CalcualteBoundingRect();
}

void CNCDrillFile::CalcualteBoundingRect()
{
    // Calculate bound rectangle of nc file
    if ( m_Coords.empty() )
    {
	m_dblXMin = 0;
	m_dblXMax = 1;
	m_dblYMin = 0;
	m_dblYMax = 1;
    }
    else
    {
	m_dblXMin = m_dblXMax = m_Coords.front()->X;
	m_dblYMin = m_dblYMax = m_Coords.front()->Y;

	std::list<NCCoord*>::iterator it;
	for ( it = m_Coords.begin(); it != m_Coords.end(); it++ )
	{
	    NCCoord &c = **it;

	    if (c.X < m_dblXMin)
		m_dblXMin = c.X;
	    else if (c.X > m_dblXMax)
		m_dblXMax = c.X;

	    if (c.Y < m_dblYMin)
		m_dblYMin = c.Y;
	    else if (c.Y > m_dblYMax)
		m_dblYMax = c.Y;
	}
    }
}
