#pragma once

// MachPCBDrill, NC drill/routing program for Mach3
// Copyright (C) 201, Frank Tkalcevic, www.franksworkshop.com

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.



class NCTool
{
private:
    int m_nId;
    double m_dblSize;

public:
    NCTool(int nId, double dblSize)
    {
	m_nId = nId;
	m_dblSize = dblSize;
    }

    double GetSize() { return m_dblSize; }
    __declspec(property(get = GetSize)) double Size;

    int GetId() { return m_nId; }
    __declspec(property(get = GetId)) int Id;
};


class NCCoord
{
protected:
    NCTool *m_tool;
    double m_x;
    double m_y;
    double m_xOrig;
    double m_yOrig;
    int m_DisplayX;
    int m_DisplayY;
    CRect m_rcDisplay;
    int m_nLine;
    bool m_bRoute;

public:
    NCCoord(NCTool *tool, double x, double y, int nLine)
	: m_tool( tool ),
	m_x( x ),
	m_y( y ),
	m_xOrig( x ),
	m_yOrig( y ),
	m_nLine( nLine),
	m_bRoute( false )
    {
    }
    virtual ~NCCoord() {}

    double GetX() { return m_x; }
    __declspec(property(get = GetX)) double X;

    double GetY() { return m_y; }
    __declspec(property(get = GetY)) double Y;

    NCTool *GetTool() { return m_tool; } 
    __declspec(property(get = GetTool)) NCTool *Tool;

    int GetDisplayX() { return m_DisplayX; }
    void SetDisplayX( int value ) { m_DisplayX = value; }
    __declspec(property(get = GetDisplayX, put=SetDisplayX)) int DisplayX;

    int GetDisplayY() { return m_DisplayY; }
    void SetDisplayY( int value ) { m_DisplayY = value; }
    __declspec(property(get = GetDisplayY, put=SetDisplayY)) int DisplayY;

    CRect &GetDisplayRect() { return m_rcDisplay; }
    __declspec(property(get = GetDisplayRect)) CRect &DisplayRect;

    int GetLine() { return m_nLine; }
    __declspec(property(get = GetLine)) int Line;

    bool GetRoute() { return m_bRoute; }
    __declspec(property(get = GetRoute)) bool Route;

    void Transform( bool bFlip, double dAngle, double dCenterX, double dCenterY )
    {
	double dFlip = bFlip ? -1 : 1;
	double a = M_PI * dAngle / 180.0;
	m_x = dCenterX + dFlip * ( m_xOrig - dCenterX ) * cos( a ) - ( m_yOrig - dCenterY ) * sin( a );
	m_y = dCenterY + dFlip * ( m_xOrig - dCenterX ) * sin( a ) + ( m_yOrig - dCenterY ) * cos( a );
    }
};

class NCRoute: public NCCoord
{
public:
    NCRoute(NCTool *tool, double x, double y, int nLine, bool bPlunged)
	: NCCoord( tool, x, y, nLine ),
	m_bPlunged( bPlunged )
    {
	m_bRoute = true;
    }

    bool GetPlunged() { return m_bPlunged; }
    __declspec(property(get = GetPlunged)) bool Plunged;
private:
    bool m_bPlunged;
};

class CNCDrillFile
{
public:
    CNCDrillFile(void);
    ~CNCDrillFile(void);

    void Load(CString sFilename);



    bool GetMetric() { return m_bMetric; }
    std::map<int,NCTool*> &GetTools() { return m_Tools; }
    std::list<NCCoord*> &GetCoords() { return m_Coords; }
    double GetXmin() { return m_dblXMin; }
    double GetXmax() { return m_dblXMax; }
    double GetYmin() { return m_dblYMin; }
    double GetYmax() { return m_dblYMax; }
    double Width() { return m_dblXMax - m_dblXMin; }
    double Height() { return m_dblYMax - m_dblYMin; }

    __declspec(property(get = GetMetric)) bool Metric;
    __declspec(property(get = GetTools)) std::map<int,NCTool*> &Tools;
    __declspec(property(get = GetCoords)) std::list<NCCoord*> &Coords;
    __declspec(property(get = GetXmin)) double Xmin;
    __declspec(property(get = GetXmax)) double Xmax;
    __declspec(property(get = GetYmin)) double Ymin;
    __declspec(property(get = GetYmax)) double Ymax;

    size_t CoordCount()
    {
	return m_Coords.size() - RouteCount();
    }
    size_t RouteCount()
    {
	int n = 0;
	std::list<NCCoord*>::iterator it = m_Coords.begin();
	for ( ; it != m_Coords.end(); it++ )
	{
	    NCRoute *r = dynamic_cast<NCRoute*>(*it);
	    if ( r != NULL )
		n++;
	}
	return n;
    }
    void CalcualteBoundingRect();

private:
    double m_dblLastX;
    double m_dblLastY;
    double m_dblXMin, m_dblXMax;
    double m_dblYMin, m_dblYMax;
    std::list<NCCoord*> m_Coords;
    std::map<int,NCTool*> m_Tools;
    bool m_bMetric;
    bool m_bLeadingZeroes;
    NCTool *m_ActiveTool;
    int m_nLineCount;
    int m_nFormat;
    bool m_bDrilling;
    bool m_bPlunged;

    void ProcessZeroes( CString &sLine );
    void ProcessToolDefintion( CString &sLine );
    double MakeDouble( CString &sLine );
    void ProcessHole( CString &sLine );
    void ProcessToolChange( CString &sLine );
    void SetFormat( CString &s );
    void ProcessCutterCompensation(CString &sLine );
    void ProcessRoute(CString &sLine );
    void ProcessGCode(CString sLine );
    void ProcessMCode(CString &sLine );
    void GetXY( CString sLine, double &x, double &y );
};

